import numpy as np
import pytest

import cyipopt


def test_hs071_solve(hs071_initial_guess_fixture, hs071_problem_instance_fixture):
	"""Test hs071 test problem solves to the correct solution."""
	x0 = hs071_initial_guess_fixture
	nlp = hs071_problem_instance_fixture
	x, info = nlp.solve(x0)

	expected_J = 17.01401714021362
	np.testing.assert_almost_equal(info["obj_val"], expected_J)

	expected_x = np.array([1.0, 4.74299964, 3.82114998, 1.37940829])
	np.testing.assert_allclose(x, expected_x)


def _make_problem(definition, lb, ub, cl, cu):
    n = len(lb)
    m = len(cl)
    return cyipopt.Problem(
        n=n, m=m, problem_obj=definition, lb=lb, ub=ub, cl=cl, cu=cu
    )


def _solve_and_assert_correct(problem, x0):
    x, info = problem.solve(x0)
    expected_x = np.array([1.0, 4.74299964, 3.82114998, 1.37940829])
    assert info["status"] == 0
    np.testing.assert_allclose(x, expected_x)


def _assert_solve_fails(problem, x0):
    x, info = problem.solve(x0)
    # The current (Ipopt 3.14.11) return status is "Invalid number".
    assert info["status"] < 0


def test_hs071_objective_eval_error(
    hs071_initial_guess_fixture,
    hs071_problem_instance_fixture,
    hs071_definition_instance_fixture,
    hs071_variable_lower_bounds_fixture,
    hs071_variable_upper_bounds_fixture,
    hs071_constraint_lower_bounds_fixture,
    hs071_constraint_upper_bounds_fixture,
):
    class ObjectiveWithError:
        def __init__(self):
            self.n_eval_error = 0

        def __call__(self, x):
            if x[0] > 1.1:
                self.n_eval_error += 1
                raise cyipopt.CyIpoptEvaluationError()
            return x[0] * x[3] * np.sum(x[0:3]) + x[2]

    objective_with_error = ObjectiveWithError()

    x0 = hs071_initial_guess_fixture
    definition = hs071_definition_instance_fixture
    definition.objective = objective_with_error
    definition.intermediate = None

    lb = hs071_variable_lower_bounds_fixture
    ub = hs071_variable_upper_bounds_fixture
    cl = hs071_constraint_lower_bounds_fixture
    cu = hs071_constraint_upper_bounds_fixture

    problem = _make_problem(definition, lb, ub, cl, cu)
    # Note that the behavior tested here (success or failure of the solve when
    # an evaluation error is raised in each callback) is documented in the
    # CyIpoptEvaluationError class. If this behavior changes (e.g. Ipopt starts
    # handling evaluation errors in the Jacobian), these tests will start to
    # fail. We will need to (a) update these tests and (b) update the
    # CyIpoptEvaluationError documentation, possibly with Ipopt version-specific
    # behavior.
    _solve_and_assert_correct(problem, x0)

    assert objective_with_error.n_eval_error > 0


def test_hs071_grad_obj_eval_error(
    hs071_initial_guess_fixture,
    hs071_problem_instance_fixture,
    hs071_definition_instance_fixture,
    hs071_variable_lower_bounds_fixture,
    hs071_variable_upper_bounds_fixture,
    hs071_constraint_lower_bounds_fixture,
    hs071_constraint_upper_bounds_fixture,
):
    class GradObjWithError:
        def __init__(self):
            self.n_eval_error = 0

        def __call__(self, x):
            if x[0] > 1.1:
                self.n_eval_error += 1
                raise cyipopt.CyIpoptEvaluationError()
            return np.array([
                x[0] * x[3] + x[3] * np.sum(x[0:3]),
                x[0] * x[3],
                x[0] * x[3] + 1.0,
                x[0] * np.sum(x[0:3]),
            ])

    gradient_with_error = GradObjWithError()

    x0 = hs071_initial_guess_fixture
    definition = hs071_definition_instance_fixture
    definition.gradient = gradient_with_error
    definition.intermediate = None

    lb = hs071_variable_lower_bounds_fixture
    ub = hs071_variable_upper_bounds_fixture
    cl = hs071_constraint_lower_bounds_fixture
    cu = hs071_constraint_upper_bounds_fixture

    problem = _make_problem(definition, lb, ub, cl, cu)
    # Solve fails when the evaluation error occurs in objective
    # gradient evaluation.
    _assert_solve_fails(problem, x0)

    # Since we fail at the first evaluation error, we know we only encountered one.
    assert gradient_with_error.n_eval_error == 1


def test_hs071_constraints_eval_error(
    hs071_initial_guess_fixture,
    hs071_problem_instance_fixture,
    hs071_definition_instance_fixture,
    hs071_variable_lower_bounds_fixture,
    hs071_variable_upper_bounds_fixture,
    hs071_constraint_lower_bounds_fixture,
    hs071_constraint_upper_bounds_fixture,
):
    class ConstraintsWithError:
        def __init__(self):
            self.n_eval_error = 0

        def __call__(self, x):
            if x[0] > 1.1:
                self.n_eval_error += 1
                raise cyipopt.CyIpoptEvaluationError()
            return np.array((np.prod(x), np.dot(x, x)))

    constraints_with_error = ConstraintsWithError()

    x0 = hs071_initial_guess_fixture
    definition = hs071_definition_instance_fixture
    definition.constraints = constraints_with_error
    definition.intermediate = None

    lb = hs071_variable_lower_bounds_fixture
    ub = hs071_variable_upper_bounds_fixture
    cl = hs071_constraint_lower_bounds_fixture
    cu = hs071_constraint_upper_bounds_fixture

    problem = _make_problem(definition, lb, ub, cl, cu)
    _solve_and_assert_correct(problem, x0)

    assert constraints_with_error.n_eval_error > 0


def test_hs071_jacobian_eval_error(
    hs071_initial_guess_fixture,
    hs071_problem_instance_fixture,
    hs071_definition_instance_fixture,
    hs071_variable_lower_bounds_fixture,
    hs071_variable_upper_bounds_fixture,
    hs071_constraint_lower_bounds_fixture,
    hs071_constraint_upper_bounds_fixture,
):
    class JacobianWithError:
        def __init__(self):
            self.n_eval_error = 0

        def __call__(self, x):
            if x[0] > 1.1:
                self.n_eval_error += 1
                raise cyipopt.CyIpoptEvaluationError()
            return np.concatenate((np.prod(x) / x, 2 * x))

    jacobian_with_error = JacobianWithError()

    x0 = hs071_initial_guess_fixture
    definition = hs071_definition_instance_fixture
    definition.jacobian = jacobian_with_error
    definition.intermediate = None

    lb = hs071_variable_lower_bounds_fixture
    ub = hs071_variable_upper_bounds_fixture
    cl = hs071_constraint_lower_bounds_fixture
    cu = hs071_constraint_upper_bounds_fixture

    problem = _make_problem(definition, lb, ub, cl, cu)
    # Solve fails when the evaluation error occurs in constraint
    # Jacobian evaluation.
    _assert_solve_fails(problem, x0)

    assert jacobian_with_error.n_eval_error == 1


def test_hs071_hessian_eval_error(
    hs071_initial_guess_fixture,
    hs071_problem_instance_fixture,
    hs071_definition_instance_fixture,
    hs071_variable_lower_bounds_fixture,
    hs071_variable_upper_bounds_fixture,
    hs071_constraint_lower_bounds_fixture,
    hs071_constraint_upper_bounds_fixture,
):
    class HessianWithError:
        def __init__(self):
            self.n_eval_error = 0

        def __call__(self, x, lagrange, obj_factor):
            if x[0] > 1.1:
                self.n_eval_error += 1
                raise cyipopt.CyIpoptEvaluationError()

            H = obj_factor * np.array([
                (2 * x[3], 0, 0, 0),
                (x[3], 0, 0, 0),
                (x[3], 0, 0, 0),
                (2 * x[0] + x[1] + x[2], x[0], x[0], 0),
            ])
            H += lagrange[0] * np.array([
                (0, 0, 0, 0),
                (x[2] * x[3], 0, 0, 0),
                (x[1] * x[3], x[0] * x[3], 0, 0),
                (x[1] * x[2], x[0] * x[2], x[0] * x[1], 0),
            ])
            H += lagrange[1] * 2 * np.eye(4)
            row, col = np.nonzero(np.tril(np.ones((4, 4))))
            return H[row, col]

    hessian_with_error = HessianWithError()

    x0 = hs071_initial_guess_fixture
    definition = hs071_definition_instance_fixture
    definition.hessian = hessian_with_error
    definition.intermediate = None

    lb = hs071_variable_lower_bounds_fixture
    ub = hs071_variable_upper_bounds_fixture
    cl = hs071_constraint_lower_bounds_fixture
    cu = hs071_constraint_upper_bounds_fixture

    problem = _make_problem(definition, lb, ub, cl, cu)
    # Solve fails when the evaluation error occurs in Lagrangian
    # Hessian evaluation.
    _assert_solve_fails(problem, x0)

    assert hessian_with_error.n_eval_error == 1



import numpy as np
import pytest

import cyipopt

pre_3_14_0 = (
    cyipopt.IPOPT_VERSION[0] < 3
    or (cyipopt.IPOPT_VERSION[0] == 3 and cyipopt.IPOPT_VERSION[1] < 14)
)


@pytest.mark.skipif(
    pre_3_14_0,
    reason="GetIpoptCurrentViolations was introduced in Ipopt v3.14.0",
    # skip these tests in old versions as the version check happens before
    # the __in_ipopt_solve check
)
def test_get_iterate_uninit(hs071_problem_instance_fixture):
    """Test that we can call get_current_iterate on an uninitialized problem
    """
    nlp = hs071_problem_instance_fixture
    msg = "can only be called during a call to solve"
    with pytest.raises(RuntimeError, match=msg):
        iterate = nlp.get_current_iterate()


@pytest.mark.skipif(
    pre_3_14_0,
    reason="GetIpoptCurrentViolations was introduced in Ipopt v3.14.0",
)
def test_get_iterate_postsolve(
    hs071_initial_guess_fixture,
    hs071_problem_instance_fixture,
):
    x0 = hs071_initial_guess_fixture
    nlp = hs071_problem_instance_fixture
    x, info = nlp.solve(x0)

    msg = "can only be called during a call to solve"
    with pytest.raises(RuntimeError, match=msg):
        iterate = nlp.get_current_iterate()
    expected_x = np.array([1.0, 4.74299964, 3.82114998, 1.37940829])
    np.testing.assert_allclose(x, expected_x)


@pytest.mark.skipif(
    pre_3_14_0,
    reason="GetIpoptCurrentViolations was introduced in Ipopt v3.14.0",
)
def test_get_violations_uninit(hs071_problem_instance_fixture):
    nlp = hs071_problem_instance_fixture
    msg = "can only be called during a call to solve"
    with pytest.raises(RuntimeError, match=msg):
        violations = nlp.get_current_violations()


@pytest.mark.skipif(
    pre_3_14_0,
    reason="GetIpoptCurrentViolations was introduced in Ipopt v3.14.0",
)
def test_get_violations_postsolve(
    hs071_initial_guess_fixture,
    hs071_problem_instance_fixture,
):
    x0 = hs071_initial_guess_fixture
    nlp = hs071_problem_instance_fixture
    x, info = nlp.solve(x0)

    msg = "can only be called during a call to solve"
    with pytest.raises(RuntimeError, match=msg):
        violations = nlp.get_current_violations()
    expected_x = np.array([1.0, 4.74299964, 3.82114998, 1.37940829])
    np.testing.assert_allclose(x, expected_x)


@pytest.mark.skipif(
    not pre_3_14_0,
    reason="GetIpoptCurrentIterate was introduced in Ipopt v3.14.0",
)
def test_get_iterate_fail_pre_3_14_0(hs071_problem_instance_fixture):
    nlp = hs071_problem_instance_fixture
    # Note that the version check happens before the __in_ipopt_solve
    # check, so we don't need to call solve to test this.
    msg = "only supports Ipopt version >=3.14.0"
    with pytest.raises(RuntimeError, match=msg):
        iterate = nlp.get_current_iterate()


@pytest.mark.skipif(
    not pre_3_14_0,
    reason="GetIpoptCurrentViolations was introduced in Ipopt v3.14.0",
)
def test_get_violations_fail_pre_3_14_0(hs071_problem_instance_fixture):
    nlp = hs071_problem_instance_fixture
    # Note that the version check happens before the __in_ipopt_solve
    # check, so we don't need to call solve to test this.
    msg = "only supports Ipopt version >=3.14.0"
    with pytest.raises(RuntimeError, match=msg):
        violations = nlp.get_current_violations()


@pytest.mark.skipif(
    pre_3_14_0,
    reason="GetIpoptCurrentIterate was introduced in Ipopt v3.14.0",
)
def test_get_iterate_hs071(
    hs071_initial_guess_fixture,
    hs071_definition_instance_fixture,
    hs071_variable_lower_bounds_fixture,
    hs071_variable_upper_bounds_fixture,
    hs071_constraint_lower_bounds_fixture,
    hs071_constraint_upper_bounds_fixture,
):
    """This test demonstrates a hacky way to call get_current_iterate from an
    intermediate callback without subclassing Problem. This is not recommended.
    """
    x0 = hs071_initial_guess_fixture
    lb = hs071_variable_lower_bounds_fixture
    ub = hs071_variable_upper_bounds_fixture
    cl = hs071_constraint_lower_bounds_fixture
    cu = hs071_constraint_upper_bounds_fixture
    n = len(x0)
    m = len(cl)

    problem_definition = hs071_definition_instance_fixture

    #
    # Define a callback that uses some "global" information to call
    # get_current_iterate and store the result
    #
    x_iterates = []
    def intermediate(
        alg_mod,
        iter_count,
        obj_value,
        inf_pr,
        inf_du,
        mu,
        d_norm,
        regularization_size,
        alpha_du,
        alpha_pr,
        ls_trials,
    ):
        # CyIpopt's C wapper expects a callback with this signature. If we
        # implemented this as a method on problem_definition, we could store
        # and access global information on self.
        # test_get_iterate_hs071_subclass_Problem tests this below.

        # This callback must be defined before constructing the Problem, but can
        # be defined after (or as part of) problem_definition. If we attach the
        # Problem to the "definition", then we can call get_current_iterate
        # from this callback.
        iterate = problem_definition.nlp.get_current_iterate(scaled=False)
        x_iterates.append(iterate["x"])

        # Hack so we may get the number of iterations after the solve
        problem_definition.iter_count = iter_count

    # Replace "intermediate" attribute with our callback, which knows
    # about the "Problem", and therefore can call get_current_iterate.
    problem_definition.intermediate = intermediate

    nlp = cyipopt.Problem(
        n=n,
        m=m,
        problem_obj=problem_definition,
        lb=lb,
        ub=ub,
        cl=cl,
        cu=cu,
    )
    # Add nlp (the "Problem") as an attribute on our "problem definition".
    # This way we can call methods on the Problem, like get_current_iterate,
    # during the solve.
    problem_definition.nlp = nlp

    # Disable bound push to make testing easier
    nlp.add_option("bound_push", 1e-9)
    x, info = nlp.solve(x0)

    # Assert correct solution
    expected_x = np.array([1.0, 4.74299964, 3.82114998, 1.37940829])
    np.testing.assert_allclose(x, expected_x)

    #
    # Assert some very basic information about the collected primal iterates
    #
    assert len(x_iterates) == (1 + problem_definition.iter_count)

    # These could be different due to bound_push (and scaling)
    np.testing.assert_allclose(x_iterates[0], x0)

    # These could be different due to honor_original_bounds (and scaling)
    np.testing.assert_allclose(x_iterates[-1], x)


@pytest.mark.skipif(
    pre_3_14_0,
    reason="GetIpoptCurrentIterate was introduced in Ipopt v3.14.0",
)
def test_get_iterate_hs071_subclass_Problem(
    hs071_initial_guess_fixture,
    hs071_definition_instance_fixture,
    hs071_variable_lower_bounds_fixture,
    hs071_variable_upper_bounds_fixture,
    hs071_constraint_lower_bounds_fixture,
    hs071_constraint_upper_bounds_fixture,
):
    x0 = hs071_initial_guess_fixture
    lb = hs071_variable_lower_bounds_fixture
    ub = hs071_variable_upper_bounds_fixture
    cl = hs071_constraint_lower_bounds_fixture
    cu = hs071_constraint_upper_bounds_fixture
    n = len(x0)
    m = len(cl)

    problem_definition = hs071_definition_instance_fixture

    x_iterates = []
    class MyProblem(cyipopt.Problem):

        def objective(self, x):
            return problem_definition.objective(x)

        def gradient(self, x):
            return problem_definition.gradient(x)

        def constraints(self, x):
            return problem_definition.constraints(x)

        def jacobian(self, x):
            return problem_definition.jacobian(x)

        def jacobian_structure(self, x):
            return problem_definition.jacobian_structure(x)

        def hessian(self, x, lagrange, obj_factor):
            return problem_definition.hessian(x, lagrange, obj_factor)

        def hessian_structure(self, x):
            return problem_definition.hessian_structure(x)

        def intermediate(
            self,
            alg_mod,
            iter_count,
            obj_value,
            inf_pr,
            inf_du,
            mu,
            d_norm,
            regularization_size,
            alpha_du,
            alpha_pr,
            ls_trials,
        ):
            # By subclassing Problem, we can call get_current_iterate
            # without any "global" information
            iterate = self.get_current_iterate(scaled=False)
            x_iterates.append(iterate["x"])

            # Hack so we may get the number of iterations after the solve
            self.iter_count = iter_count

    nlp = MyProblem(
        n=n,
        m=m,
        lb=lb,
        ub=ub,
        cl=cl,
        cu=cu,
    )

    # Disable bound push to make testing easier
    nlp.add_option("bound_push", 1e-9)
    x, info = nlp.solve(x0)

    # Assert correct solution
    expected_x = np.array([1.0, 4.74299964, 3.82114998, 1.37940829])
    np.testing.assert_allclose(x, expected_x)

    #
    # Assert some very basic information about the collected primal iterates
    #
    assert len(x_iterates) == (1 + nlp.iter_count)

    # These could be different due to bound_push (and scaling)
    np.testing.assert_allclose(x_iterates[0], x0)

    # These could be different due to honor_original_bounds (and scaling)
    np.testing.assert_allclose(x_iterates[-1], x)


@pytest.mark.skipif(
    pre_3_14_0,
    reason="GetIpoptCurrentViolations was introduced in Ipopt v3.14.0",
)
def test_get_violations_hs071_subclass_Problem(
    hs071_initial_guess_fixture,
    hs071_definition_instance_fixture,
    hs071_variable_lower_bounds_fixture,
    hs071_variable_upper_bounds_fixture,
    hs071_constraint_lower_bounds_fixture,
    hs071_constraint_upper_bounds_fixture,
):
    x0 = hs071_initial_guess_fixture
    lb = hs071_variable_lower_bounds_fixture
    ub = hs071_variable_upper_bounds_fixture
    cl = hs071_constraint_lower_bounds_fixture
    cu = hs071_constraint_upper_bounds_fixture
    n = len(x0)
    m = len(cl)

    problem_definition = hs071_definition_instance_fixture

    pr_violations = []
    du_violations = []
    iter_counts = []
    class MyProblem(cyipopt.Problem):

        def objective(self, x):
            return problem_definition.objective(x)

        def gradient(self, x):
            return problem_definition.gradient(x)

        def constraints(self, x):
            return problem_definition.constraints(x)

        def jacobian(self, x):
            return problem_definition.jacobian(x)

        def jacobian_structure(self, x):
            return problem_definition.jacobian_structure(x)

        def hessian(self, x, lagrange, obj_factor):
            return problem_definition.hessian(x, lagrange, obj_factor)

        def hessian_structure(self, x):
            return problem_definition.hessian_structure(x)

        def intermediate(
            self,
            alg_mod,
            iter_count,
            obj_value,
            inf_pr,
            inf_du,
            mu,
            d_norm,
            regularization_size,
            alpha_du,
            alpha_pr,
            ls_trials,
        ):
            violations = self.get_current_violations(scaled=True)
            pr_violations.append(violations["g_violation"])
            du_violations.append(violations["grad_lag_x"])

            # Hack so we may get the number of iterations after the solve
            iter_counts.append(iter_count)

    nlp = MyProblem(
        n=n,
        m=m,
        lb=lb,
        ub=ub,
        cl=cl,
        cu=cu,
    )

    nlp.add_option("tol", 1e-8)
    # Note that Ipopt appears to check tolerance in the scaled, bound-relaxed
    # NLP. To ensure our intermediate infeasibilities, which are in the user's
    # original NLP, are less than the above tolerance at the final iteration,
    # we must turn off bound relaxation.
    nlp.add_option("bound_relax_factor", 0.0)
    x, info = nlp.solve(x0)

    # Assert correct solution
    expected_x = np.array([1.0, 4.74299964, 3.82114998, 1.37940829])
    np.testing.assert_allclose(x, expected_x)

    #
    # Assert some very basic information about the collected violations
    #
    iter_count = iter_counts[-1]
    assert len(pr_violations) == (1 + iter_count)
    assert len(du_violations) == (1 + iter_count)

    np.testing.assert_allclose(pr_violations[-1], np.zeros(m), atol=1e-8)
    np.testing.assert_allclose(du_violations[-1], np.zeros(n), atol=1e-8)



# -*- coding: utf-8 -*-
"""
cyipopt: Python wrapper for the Ipopt optimization package, written in Cython.

Copyright (C) 2012-2015 Amit Aides
Copyright (C) 2015-2017 Matthias Kümmerer
Copyright (C) 2017-2024 cyipopt developers

License: EPL 2.0

Usage::

   $ python lasso.py
   $ python lasso.py -p
   $ python lasso.py --plot
"""
#
# This function executes IPOPT to find the maximum likelihood solution to
# least squares regression with L1 regularization or the "Lasso". The inputs
# are the data matrix A (in which each row is an example vector), the vector
# of regression outputs y, and the penalty parameter lambda, a number
# greater than zero. The output is the estimated vector of regression
# coefficients.
#
# Adapted from matlab code by Peter Carbonetto.
#
import numpy as np
import cyipopt


class lasso(cyipopt.Problem):
    def __init__(self, A, y):

        self._A = A
        self._y = y
        self._m = A.shape[1]

        #
        # The constraint functions are bounded from below by zero.
        #
        cl = np.zeros(2*self._m)

        super(lasso, self).__init__(
                            2*self._m,
                            2*self._m,
                            cl=cl
                            )

        #
        # Set solver options
        #
        self.add_option('derivative_test', 'second-order')
        self.add_option('jac_d_constant', 'yes')
        self.add_option('hessian_constant', 'yes')
        self.add_option('mu_strategy', 'adaptive')
        self.add_option('max_iter', 100)
        self.add_option('tol', 1e-8)

    def solve(self, _lambda):

        x0 = np.concatenate((np.zeros(m), np.ones(m)))
        self._lambda = _lambda
        x, info = super(lasso, self).solve(x0)

        return x[:self._m]

    def objective(self, x):

        w = x[:self._m].reshape((-1, 1))
        u = x[self._m:].reshape((-1, 1))

        return np.linalg.norm(self._y - np.dot(self._A, w))**2/2 + self._lambda * np.sum(u)

    def constraints(self, x):

        w = x[:self._m].reshape((-1, 1))
        u = x[self._m:].reshape((-1, 1))

        return np.vstack((u + w,  u - w))

    def gradient(self, x):

        w = x[:self._m].reshape((-1, 1))

        g = np.vstack((np.dot(-self._A.T, self._y - np.dot(self._A, w)), self._lambda*np.ones((self._m, 1))))

        return g

    def jacobianstructure(self):

        #
        # Create a sparse matrix to hold the jacobian structure
        #
        return np.nonzero(np.tile(np.eye(self._m), (2, 2)))

    def jacobian(self, x):

        I = np.eye(self._m)

        J = np.vstack((np.hstack((I, I)), np.hstack((-I, I))))

        row, col = self.jacobianstructure()

        return J[row, col]

    def hessianstructure(self):

        h = np.zeros((2*self._m, 2*self._m))
        h[:self._m, :self._m] = np.tril(np.ones((self._m, self._m)))

        #
        # Create a sparse matrix to hold the hessian structure
        #
        return np.nonzero(h)

    def hessian(self, x, lagrange, obj_factor):

        H = np.zeros((2*self._m, 2*self._m))
        H[:self._m, :self._m] = np.tril(np.tril(np.dot(self._A.T, self._A)))

        row, col = self.hessianstructure()

        return obj_factor*H[row, col]


if __name__ == '__main__':

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-p','--plot', help='Plot results with matplotlib',
                        action='store_true')
    args = parser.parse_args()


    #
    # _lambda - Level of L1 regularization
    # n - Number of training examples
    # e - Std. dev. in noise of outputs
    # beta - "True" regression coefficients
    #
    n = 100
    e = 1
    beta = np.array((0, 0, 2, -4, 0, 0, -1, 3), dtype=np.float64).reshape((-1, 1))

    #
    # Set the random number generator seed.
    #
    seed = 7
    np.random.seed(seed)

    #
    # CREATE DATA SET.
    # Generate the input vectors from the standard normal, and generate the
    # binary responses from the regression with some additional noise, and then
    # transform the results using the logistic function. The variable "beta" is
    # the set of true regression coefficients.
    #
    # A - The n x m matrix of examples
    # noise - Noise in outputs
    # y - The binary outputs
    #
    m = len(beta)
    A = np.random.randn(n, m)
    noise = e * np.random.randn(n, 1)
    y = np.dot(A, beta) + noise

    #
    # COMPUTE SOLUTION WITH IPOPT.
    # Compute the L1-regularized maximum likelihood estimator.
    #
    problem = lasso(A, y)

    LAMBDA_SAMPLES = 30
    _lambdas = np.logspace(0, 2.6, LAMBDA_SAMPLES)
    estim_betas = np.zeros((LAMBDA_SAMPLES, beta.size))

    for i, _lambda in enumerate(_lambdas):
        estim_betas[i, :] = problem.solve(_lambda)

    if args.plot:
        import matplotlib.pyplot as plt
        plt.plot(_lambdas, estim_betas)
        plt.xlabel('L1 Regularization Parameter')
        plt.ylabel('Estimated Regression Coefficients')
        plt.title('Lasso Solution as Function of The Regularization Parameter')
        plt.show()



"""Test functionality of CyIpopt with/without optional SciPy dependency.

SciPy is an optional dependency of CyIpopt. CyIpopt needs to function without
SciPy installed, but also needs to provide the :func:`minimize_ipopt` function
which requires SciPy.

"""

import re
import sys

import numpy as np
import pytest

import cyipopt

# Hard-code rather than importing from scipy.optimize._minimize in a try/except
try:
    from scipy.optimize._minimize import MINIMIZE_METHODS
except ImportError:
    MINIMIZE_METHODS = []

@pytest.mark.skipif("scipy" in sys.modules,
                    reason="Test only valid if no Scipy available.")
def test_minimize_ipopt_import_error_if_no_scipy():
    """`minimize_ipopt` not callable without SciPy installed."""
    expected_error_msg = re.escape("Install SciPy to use the "
                                   "`minimize_ipopt` function.")
    with pytest.raises(ImportError, match=expected_error_msg):
        cyipopt.minimize_ipopt(None, None)


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
def test_tol_type_issue_235():
    # from: https://github.com/mechmotum/cyipopt/issues/235

    def fun(x):
        return np.sum(x ** 2)

    # tol should not raise an error
    cyipopt.minimize_ipopt(
        fun=fun,
        x0=np.zeros(2),
        tol=1e-9,
    )


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
def test_minimize_ipopt_input_validation():
    from scipy import optimize

    x0 = 1
    def f(x):
        return x @ x

    message = "`fun` must be callable."
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt('migratory coconuts', x0)

    message = "`x0` must be a numeric array."
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, 'spamalot')

    message = "`kwargs` must be a dictionary."
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, x0, kwargs='elderberries')

    message = "Unknown solver..."
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, x0, method='a newt')

    message = "`jac` must be callable or boolean."
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, x0, jac='self-perpetuating autocracy')

    message = "`hess` must be callable."
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, x0, hess='farcical aquatic ceremony')

    message = "`hessp` is not yet supported by Ipopt."
    with pytest.raises(NotImplementedError, match=message):
        cyipopt.minimize_ipopt(f, x0, hessp='shrubbery')

    message = "`bounds` must specify both lower and upper..."
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, [1, 2], bounds=1)

    message = "The number of lower bounds, upper bounds..."
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, [1, 2], bounds=[(1, 2), (3, 4), (5, 6)])

    message = "The bounds must be numeric."
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, x0, bounds=[['low', 'high']])

    message = "`callback` is not yet supported by Ipopt."
    with pytest.raises(NotImplementedError, match=message):
        cyipopt.minimize_ipopt(f, x0, callback='a duck')

    message = "`tol` must be a positive scalar."
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, x0, tol=[1, 2, 3])
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, x0, tol='ni')
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, x0, tol=-1)

    message = "`options` must be a dictionary."
    with pytest.raises(ValueError, match=message):
        cyipopt.minimize_ipopt(f, x0, options='supreme executive power')


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
def test_minimize_ipopt_if_scipy():
    """If SciPy is installed `minimize_ipopt` works correctly."""
    from scipy.optimize import rosen, rosen_der
    x0 = [1.3, 0.7, 0.8, 1.9, 1.2]
    res = cyipopt.minimize_ipopt(rosen, x0, jac=rosen_der)
    assert isinstance(res, dict)
    assert np.isclose(res.get("fun"), 0.0)
    assert res.get("status") == 0
    assert res.get("success") is True
    np.testing.assert_allclose(res.get("x"), np.ones(5))


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
def test_minimize_ipopt_nojac_if_scipy():
    """`minimize_ipopt` works without Jacobian."""
    from scipy.optimize import rosen
    x0 = [1.3, 0.7, 0.8, 1.9, 1.2]
    options = {"tol": 1e-7}
    res = cyipopt.minimize_ipopt(rosen, x0, options=options)

    assert isinstance(res, dict)
    assert np.isclose(res.get("fun"), 0.0)
    assert res.get("status") == 0
    assert res.get("success") is True
    np.testing.assert_allclose(res.get("x"), np.ones(5), rtol=1e-5)


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
def test_minimize_ipopt_nojac_constraints_if_scipy():
    """ `minimize_ipopt` works without Jacobian and with constraints"""
    from scipy.optimize import rosen
    x0 = [1.3, 0.7, 0.8, 1.9, 1.2]
    constr = {"fun": lambda x: rosen(x) - 1.0, "type": "ineq"}
    res = cyipopt.minimize_ipopt(rosen, x0, constraints=constr)
    assert isinstance(res, dict)
    assert np.isclose(res.get("fun"), 1.0)
    assert res.get("status") == 0
    assert res.get("success") is True
    expected_res = np.array([1.001867, 0.99434067, 1.05070075, 1.17906312,
                             1.38103001])
    np.testing.assert_allclose(res.get("x"), expected_res)


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
def test_minimize_ipopt_jac_and_hessians_constraints_if_scipy(
):
    """`minimize_ipopt` works with objective gradient and Hessian
       and constraint jacobians and Hessians."""
    from scipy.optimize import rosen, rosen_der, rosen_hess
    x0 = [0.0, 0.0]
    constr = {
        "type": "ineq",
        "fun": lambda x: -x[0]**2 - x[1]**2 + 2,
        "jac": lambda x: np.array([-2 * x[0], -2 * x[1]]),
        "hess": lambda x, v: -2 * np.eye(2) * v[0]
    }
    bounds = [(-1.5, 1.5), (-1.5, 1.5)]
    res = cyipopt.minimize_ipopt(rosen, x0, jac=rosen_der, hess=rosen_hess,
                                 constraints=constr)
    assert isinstance(res, dict)
    assert np.isclose(res.get("fun"), 0.0)
    assert res.get("status") == 0
    assert res.get("success") is True
    expected_res = np.array([1.0, 1.0])
    np.testing.assert_allclose(res.get("x"), expected_res, rtol=1e-5)


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
def test_minimize_ipopt_jac_hessians_constraints_with_arg_kwargs():
    """Makes sure that args and kwargs can be passed to all user defined
    functions in minimize_ipopt."""
    from scipy.optimize import rosen, rosen_der, rosen_hess

    rosen2 = lambda x, a, b=None: rosen(x)*a*b
    rosen_der2 = lambda x, a, b=None: rosen_der(x)*a*b
    rosen_hess2 = lambda x, a, b=None: rosen_hess(x)*a*b

    x0 = [0.0, 0.0]
    constr = {
        "type": "ineq",
        "fun": lambda x, a, b=None: -x[0]**2 - x[1]**2 + 2*a*b,
        "jac": lambda x, a, b=None: np.array([-2 * x[0], -2 * x[1]])*a*b,
        "hess": lambda x, v, a, b=None: -2 * np.eye(2) * v[0]*a*b,
        "args": (1.0, ),
        "kwargs": {'b': 1.0},
    }
    res = cyipopt.minimize_ipopt(rosen2, x0,
                                 jac=rosen_der2,
                                 hess=rosen_hess2,
                                 args=constr['args'],
                                 kwargs=constr['kwargs'],
                                 constraints=constr)
    assert isinstance(res, dict)
    assert np.isclose(res.get("fun"), 0.0)
    assert res.get("status") == 0
    assert res.get("success") is True
    expected_res = np.array([1.0, 1.0])
    np.testing.assert_allclose(res.get("x"), expected_res, rtol=1e-5)


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
@pytest.mark.parametrize('method', [None] + MINIMIZE_METHODS)
def test_minimize_ipopt_jac_with_scipy_methods(method):
    x0 = [0] * 4
    a0, b0, c0, d0 = 1, 2, 3, 4
    atol, rtol = 5e-5, 5e-5

    def fun(x, a=0, e=0, b=0):
        assert a == a0
        assert b == b0
        fun.count += 1
        return (x[0] - a) ** 2 + (x[1] - b) ** 2 + x[2] ** 2 + x[3] ** 2

    def grad(x, a=0, e=0, b=0):
        assert a == a0
        assert b == b0
        grad.count += 1
        return [2 * (x[0] - a), 2 * (x[1] - b), 2 * x[2], 2 * x[3]]

    def hess(x, a=0, e=0, b=0):
        assert a == a0
        assert b == b0
        hess.count += 1
        return 2 * np.eye(4)

    def hessp(x, p, a=0, e=0, b=0):
        assert a == a0
        assert b == b0
        hessp.count += 1
        return 2 * np.eye(4) @ p

    def fun_constraint(x, c=0, e=0, d=0):
        assert c == c0
        assert d == d0
        fun_constraint.count += 1
        return [(x[2] - c) ** 2, (x[3] - d) ** 2]

    def grad_constraint(x, c=0, e=0, d=0):
        assert c == c0
        assert d == d0
        grad_constraint.count += 1
        return np.hstack((np.zeros((2, 2)),
                         np.diag([2 * (x[2] - c), 2 * (x[3] - d)])))

    def callback(*args, **kwargs):
        callback.count += 1

    constr = {
        "type": "eq",
        "fun": fun_constraint,
        "jac": grad_constraint,
        "args": (c0,),
        "kwargs": {'d': d0},
    }

    kwargs = {}
    jac_methods = {'cg', 'bfgs', 'newton-cg', 'l-bfgs-b', 'tnc', 'slsqp,',
                   'dogleg', 'trust-ncg', 'trust-krylov', 'trust-exact',
                   'trust-constr'}
    hess_methods = {'newton-cg', 'dogleg', 'trust-ncg', 'trust-krylov',
                   'trust-exact', 'trust-constr'}
    hessp_methods = hess_methods - {'dogleg', 'trust-exact'}
    constr_methods = {'slsqp', 'trust-constr'}

    if method in jac_methods:
        kwargs['jac'] = grad
    if method in hess_methods:
        kwargs['hess'] = hess
    if method in constr_methods:
        kwargs['constraints'] = constr
    if method in MINIMIZE_METHODS:
        kwargs['callback'] = callback

    fun.count = 0
    grad.count = 0
    hess.count = 0
    hessp.count = 0
    fun_constraint.count = 0
    grad_constraint.count = 0
    callback.count = 0

    res = cyipopt.minimize_ipopt(fun, x0, method=method, args=(a0,),
                                 kwargs={'b': b0}, **kwargs)

    assert res.success
    np.testing.assert_allclose(res.x[:2], [a0, b0], rtol=rtol)

    # confirm that the test covers what we think it does: all the functions
    # that we provide are actually being executed; that is, the assertions
    # are *passing*, not being skipped
    assert fun.count > 0
    if method in MINIMIZE_METHODS:
        assert callback.count > 0
    if method in jac_methods:
        assert grad.count > 0
    if method in hess_methods:
        assert hess.count > 0
    if method in constr_methods:
        assert fun_constraint.count > 0
        assert grad_constraint.count > 0
        np.testing.assert_allclose(res.x[2:], [c0, d0], rtol=rtol)
    else:
        np.testing.assert_allclose(res.x[2:], 0, atol=atol)

    # For methods that support `hessp`, check that it works, too.
    if method in hessp_methods:
        del kwargs['hess']
        kwargs['hessp'] = hessp

        res = cyipopt.minimize_ipopt(fun, x0, method=method, args=(a0,),
                                     kwargs={'b': b0}, **kwargs)
        assert res.success
        assert hessp.count > 0
        np.testing.assert_allclose(res.x[:2], [a0, b0], rtol=rtol)
        if method in constr_methods:
            np.testing.assert_allclose(res.x[2:], [c0, d0], rtol=rtol)
        else:
            np.testing.assert_allclose(res.x[2:], 0, atol=atol)


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid of Scipy available")
def test_minimize_ipopt_bounds_tol_options():
    # spot check additional cases not tested above
    def fun(x):
        return x**2

    x0 = 2.

    # make sure `bounds` is passed to SciPy methods
    bounds = (1, 3)
    res = cyipopt.minimize_ipopt(fun, x0, method='slsqp', bounds=[(1, 3)])
    np.testing.assert_allclose(res.x, bounds[0])

    # make sure `tol` is passed to SciPy methods
    tol1, tol2 = 1e-3, 1e-9
    res1 = cyipopt.minimize_ipopt(fun, x0, method='cobyla', tol=tol1)
    res2 = cyipopt.minimize_ipopt(fun, x0, method='cobyla', tol=tol2)
    assert abs(res2.x[0]) <= tol2 < abs(res1.x[0]) <= tol1

    # make sure `options` is passed to SciPy methods
    res = cyipopt.minimize_ipopt(fun, x0, method='slsqp')
    assert res.nit > 1
    options = dict(maxiter=1)
    res = cyipopt.minimize_ipopt(fun, x0, method='slsqp', options=options)
    assert res.nit == 1


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid of Scipy available")
def test_minimize_ipopt_sparse_jac_if_scipy():
    """ `minimize_ipopt` works with objective gradient, and sparse
        constraint jacobian. Solves
        Hock & Schittkowski's test problem 71:

        min x0*x3*(x0+x1+x2)+x2
        s.t. x0**2 + x1**2 + x2**2 + x3**2 - 40  = 0
                         x0 * x1 * x2 * x3 - 25 >= 0
                               1 <= x0,x1,x2,x3 <= 5
    """
    try:
        from scipy.sparse import coo_array
    except ImportError:
        from scipy.sparse import coo_matrix as coo_array

    def obj(x):
        return x[0] * x[3] * np.sum(x[:3]) + x[2]

    def grad(x):
        return np.array([
            x[0] * x[3] + x[3] * np.sum(x[0:3]), x[0] * x[3],
            x[0] * x[3] + 1.0, x[0] * np.sum(x[0:3])
        ])

    # Note:
    # coo_array(dense_jac_val(x)) only works if dense_jac_val(x0)
    # doesn't contain any zeros for the initial guess x0

    con_eq = {
        "type": "eq",
        "fun": lambda x: np.sum(x**2) - 40,
        "jac": lambda x: coo_array(2 * x)
    }
    con_ineq = {
        "type": "ineq",
        "fun": lambda x: np.prod(x) - 25,
        "jac": lambda x: coo_array(np.prod(x) / x),
    }
    constrs = (con_eq, con_ineq)

    x0 = np.array([1.0, 5.0, 5.0, 1.0])
    bnds = [(1, 5) for _ in range(x0.size)]

    res = cyipopt.minimize_ipopt(obj, jac=grad, x0=x0,
                                 bounds=bnds, constraints=constrs)
    assert isinstance(res, dict)
    assert np.isclose(res.get("fun"), 17.01401727277449)
    assert res.get("status") == 0
    assert res.get("success") is True
    expected_res = np.array([0.99999999, 4.74299964, 3.82114998, 1.3794083])
    np.testing.assert_allclose(res.get("x"), expected_res)


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid of Scipy available")
def test_minimize_ipopt_sparse_and_dense_jac_if_scipy():
    """ `minimize_ipopt` works with objective gradient, and sparse
        constraint jacobian. Solves
        Hock & Schittkowski's test problem 71:

        min x0*x3*(x0+x1+x2)+x2
        s.t. x0**2 + x1**2 + x2**2 + x3**2 - 40  = 0
                         x0 * x1 * x2 * x3 - 25 >= 0
                               1 <= x0,x1,x2,x3 <= 5
    """
    try:
        from scipy.sparse import coo_array
    except ImportError:
        from scipy.sparse import coo_matrix as coo_array

    def obj(x):
        return x[0] * x[3] * np.sum(x[:3]) + x[2]

    def grad(x):
        return np.array([
            x[0] * x[3] + x[3] * np.sum(x[0:3]), x[0] * x[3],
            x[0] * x[3] + 1.0, x[0] * np.sum(x[0:3])
        ])

    # Note:
    # coo_array(dense_jac_val(x)) only works if dense_jac_val(x0)
    # doesn't contain any zeros for the initial guess x0

    con_eq_dense = {
        "type": "eq",
        "fun": lambda x: np.sum(x**2) - 40,
        "jac": lambda x: 2 * x
    }
    con_ineq_sparse = {
        "type": "ineq",
        "fun": lambda x: np.prod(x) - 25,
        "jac": lambda x: coo_array(np.prod(x) / x),
    }
    constrs = (con_eq_dense, con_ineq_sparse)

    x0 = np.array([1.0, 5.0, 5.0, 1.0])
    bnds = [(1, 5) for _ in range(x0.size)]

    res = cyipopt.minimize_ipopt(obj, jac=grad, x0=x0,
                                 bounds=bnds, constraints=constrs)
    assert isinstance(res, dict)
    assert np.isclose(res.get("fun"), 17.01401727277449)
    assert res.get("status") == 0
    assert res.get("success") is True
    expected_res = np.array([0.99999999, 4.74299964, 3.82114998, 1.3794083])
    np.testing.assert_allclose(res.get("x"), expected_res)


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
def test_minimize_ipopt_hs071():
    """ `minimize_ipopt` works with objective gradient and Hessian and
         constraint jacobians and Hessians.

        The objective and the constraints functions return a tuple containing
        the function value and the evaluated gradient or jacobian. Solves
        Hock & Schittkowski's test problem 71:

        min x0*x3*(x0+x1+x2)+x2
        s.t. x0**2 + x1**2 + x2**2 + x3**2 - 40  = 0
                         x0 * x1 * x2 * x3 - 25 >= 0
                               1 <= x0,x1,x2,x3 <= 5
    """

    def obj_and_grad(x):
        obj = x[0] * x[3] * np.sum(x[:3]) + x[2]
        grad = np.array([
            x[0] * x[3] + x[3] * np.sum(x[0:3]), x[0] * x[3],
            x[0] * x[3] + 1.0, x[0] * np.sum(x[0:3])
        ])
        return obj, grad

    def obj_hess(x):
        return np.array([[2 * x[3], 0.0, 0, 0], [x[3], 0, 0, 0],
                         [x[3], 0, 0, 0],
                         [2 * x[0] + x[1] + x[2], x[0], x[0], 0]])

    def con_eq_and_jac(x):
        value = np.sum(x**2) - 40
        jac = np.array([2 * x])
        return value, jac

    def con_eq_hess(x, v):
        return v[0] * 2.0 * np.eye(4)

    def con_ineq_and_jac(x):
        value = np.prod(x) - 25
        jac = np.array([np.prod(x) / x])
        return value, jac

    def con_ineq_hess(x, v):
        return v[0] * np.array([[0, 0, 0, 0], [x[2] * x[3], 0, 0, 0],
                                [x[1] * x[3], x[0] * x[3], 0, 0],
                                [x[1] * x[2], x[0] * x[2], x[0] * x[1], 0]])

    con1 = {
        "type": "eq",
        "fun": con_eq_and_jac,
        "jac": True,
        "hess": con_eq_hess
    }
    con2 = {
        "type": "ineq",
        "fun": con_ineq_and_jac,
        "jac": True,
        "hess": con_ineq_hess
    }
    constrs = (con1, con2)

    x0 = np.array([1.0, 5.0, 5.0, 1.0])
    bnds = [(1, 5) for _ in range(x0.size)]

    res = cyipopt.minimize_ipopt(obj_and_grad, jac=True, hess=obj_hess, x0=x0,
                                 bounds=bnds, constraints=constrs)
    assert isinstance(res, dict)
    assert np.isclose(res.get("fun"), 17.01401727277449)
    assert res.get("status") == 0
    assert res.get("success") is True
    expected_res = np.array([0.99999999, 4.74299964, 3.82114998, 1.3794083])
    np.testing.assert_allclose(res.get("x"), expected_res)


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
def test_minimize_ipopt_bounds():
    # Test that `minimize_ipopt` accepts  bounds sequences or `optimize.Bounds`
    from scipy import optimize

    def f(x):
        return x @ x

    # accept size 2 sequence containing same bounds for all variables
    res = cyipopt.minimize_ipopt(f, [2, 3], bounds=[1, 10])
    np.testing.assert_allclose(res.x, [1, 1])

    res = cyipopt.minimize_ipopt(f, [2, 3], bounds=[None, None])
    np.testing.assert_allclose(res.x, [0, 0], atol=1e-6)

    # accept instance of Bounds
    bounds = optimize.Bounds(lb=0.5, ub=[1, 2])
    res = cyipopt.minimize_ipopt(f, [2, 3], bounds=bounds)
    np.testing.assert_allclose(res.x, [0.5, 0.5])


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
def test_minimize_late_binding_bug():
    # `IpoptProblemWrapper` had a late binding bug when constraint Jacobians
    # were defined with `optimize.approx_fprime`. Check that this is resolved.
    from scipy.optimize import minimize

    fun = lambda x: (x[0] - 1)**2 + (x[1] - 2.5)**2
    cons = ({'type': 'ineq', 'fun': lambda x:  x[0] - 2 * x[1] + 2},
            {'type': 'ineq', 'fun': lambda x: -x[0] - 2 * x[1] + 6},
            {'type': 'ineq', 'fun': lambda x: -x[0] + 2 * x[1] + 2})
    bnds = ((0, None), (0, None))

    res = cyipopt.minimize_ipopt(fun, (2, 0), bounds=bnds, constraints=cons)
    ref = minimize(fun, (2, 0), bounds=bnds, constraints=cons)
    assert res.success
    np.testing.assert_allclose(res.x, ref.x)
    np.testing.assert_allclose(res.fun, ref.fun)


@pytest.mark.skipif("scipy" not in sys.modules,
                    reason="Test only valid if Scipy available.")
def test_gh115_eps_option():
    # gh-115 requested that the `eps` argument be exposed as an option. Verify
    # that it is working as advertised (at least for the objective function).
    def f(x):
        if f.x is None:
            f.x = x
        elif f.dx is None:
            f.dx = x - f.x
        return x ** 2

    f.x, f.dx = None, None
    cyipopt.minimize_ipopt(f, x0=0)
    np.testing.assert_equal(f.dx, 1e-8)

    f.x, f.dx = None, None
    eps = 1e-9
    cyipopt.minimize_ipopt(f, x0=0, options={'eps': eps})
    np.testing.assert_equal(f.dx, eps)



from jax.config import config

# Enable 64 bit floating point precision
config.update("jax_enable_x64", True)

# We use the CPU instead of GPU und mute all warnings if no GPU/TPU is found.
config.update('jax_platform_name', 'cpu')

from cyipopt import minimize_ipopt
from jax import jit, grad, jacrev, jacfwd
import jax.numpy as np


# Test the scipy interface on the Hock & Schittkowski test problem 71:
#
# min x0*x3*(x0+x1+x2)+x2
#
# s.t. x0**2 + x1**2 + x2**2 + x3**2 - 40 = 0
#      x0 * x1 * x2 * x3 - 25 >= 0
#      1 <= x0,x1,x2,x3 <= 5
#
# We evaluate all derivatives (except the Hessian) by algorithmic differentation
# by means of the JAX library.


def objective(x):
    return x[0]*x[3]*np.sum(x[:3]) + x[2]


def eq_constraints(x):
    return np.sum(x**2) - 40


def ineq_constrains(x):
    return np.prod(x) - 25


# jit the functions
obj_jit = jit(objective)
con_eq_jit = jit(eq_constraints)
con_ineq_jit = jit(ineq_constrains)

# build the derivatives and jit them
obj_grad = jit(grad(obj_jit))  # objective gradient
obj_hess = jit(jacrev(jacfwd(obj_jit)))  # objective hessian
con_eq_jac = jit(jacfwd(con_eq_jit))  # jacobian
con_ineq_jac = jit(jacfwd(con_ineq_jit))  # jacobian
con_eq_hess = jacrev(jacfwd(con_eq_jit)) # hessian
con_eq_hessvp = jit(lambda x, v: con_eq_hess(x) * v[0]) # hessian vector-product
con_ineq_hess = jacrev(jacfwd(con_ineq_jit))  # hessian
con_ineq_hessvp = jit(lambda x, v: con_ineq_hess(x) * v[0])  # hessian vector-product

# constraints
# Note that 'hess' is the hessian-vector-product
cons = [
    {'type': 'eq', 'fun': con_eq_jit, 'jac': con_eq_jac, 'hess': con_eq_hessvp},
    {'type': 'ineq', 'fun': con_ineq_jit, 'jac': con_ineq_jac, 'hess': con_ineq_hessvp},
]

# initial guess
x0 = np.array([1.0, 5.0, 5.0, 1.0])

# variable bounds: 1 <= x[i] <= 5
bnds = [(1, 5) for _ in range(x0.size)]

res = minimize_ipopt(obj_jit, jac=obj_grad, hess=obj_hess, x0=x0, bounds=bnds,
                     constraints=cons, options={'disp': 5})

print(res)



# -*- coding: utf-8 -*-
"""
cyipopt: Python wrapper for the Ipopt optimization package, written in Cython.

Copyright (C) 2012-2015 Amit Aides
Copyright (C) 2015-2017 Matthias Kümmerer
Copyright (C) 2017-2024 cyipopt developers

License: EPL 2.0
"""

# Test the "ipopt" Python interface on the Hock & Schittkowski test problem
# #71. See: Willi Hock and Klaus Schittkowski. (1981) Test Examples for
# Nonlinear Programming Codes. Lecture Notes in Economics and Mathematical
# Systems Vol. 187, Springer-Verlag.
#
# Based on matlab code by Peter Carbonetto.

import numpy as np
import cyipopt


def explode():
    raise ValueError('bum')

class hs071(object):
    def __init__(self):
        self.counter = 0

    def objective(self, x):
        #
        # The callback for calculating the objective
        #
        self.counter += 1
        if self.counter > 3:
            explode()
        return x[0] * x[3] * np.sum(x[0:3]) + x[2]

    def gradient(self, x):
        #
        # The callback for calculating the gradient
        #
        return np.array([
                    x[0] * x[3] + x[3] * np.sum(x[0:3]),
                    x[0] * x[3],
                    x[0] * x[3] + 1.0,
                    x[0] * np.sum(x[0:3])
                    ])

    def constraints(self, x):
        #
        # The callback for calculating the constraints
        #
        return np.array((np.prod(x), np.dot(x, x)))

    def jacobian(self, x):
        #
        # The callback for calculating the Jacobian
        #
        return np.concatenate((np.prod(x) / x, 2*x))

    def hessianstructure(self):
        #
        # The structure of the Hessian
        # Note:
        # The default hessian structure is of a lower triangular matrix. Therefore
        # this function is redundant. I include it as an example for structure
        # callback.
        #

        return np.nonzero(np.tril(np.ones((4, 4))))

    def hessian(self, x, lagrange, obj_factor):
        #
        # The callback for calculating the Hessian
        #
        H = obj_factor*np.array((
                (2*x[3], 0, 0, 0),
                (x[3],   0, 0, 0),
                (x[3],   0, 0, 0),
                (2*x[0]+x[1]+x[2], x[0], x[0], 0)))

        H += lagrange[0]*np.array((
                (0, 0, 0, 0),
                (x[2]*x[3], 0, 0, 0),
                (x[1]*x[3], x[0]*x[3], 0, 0),
                (x[1]*x[2], x[0]*x[2], x[0]*x[1], 0)))

        H += lagrange[1]*2*np.eye(4)

        row, col = self.hessianstructure()

        return H[row, col]

    def intermediate(
            self,
            alg_mod,
            iter_count,
            obj_value,
            inf_pr,
            inf_du,
            mu,
            d_norm,
            regularization_size,
            alpha_du,
            alpha_pr,
            ls_trials
            ):

        #
        # Example for the use of the intermediate callback.
        #
        print("Objective value at iteration #%d is - %g" % (iter_count, obj_value))


def main():
    #
    # Define the problem
    #
    x0 = [1.0, 5.0, 5.0, 1.0]

    lb = [1.0, 1.0, 1.0, 1.0]
    ub = [5.0, 5.0, 5.0, 5.0]

    cl = [25.0, 40.0]
    cu = [2.0e19, 40.0]

    nlp = cyipopt.Problem(
                n=len(x0),
                m=len(cl),
                problem_obj=hs071(),
                lb=lb,
                ub=ub,
                cl=cl,
                cu=cu
                )

    #
    # Set solver options
    #
    #nlp.addOption('derivative_test', 'second-order')
    nlp.add_option('mu_strategy', 'adaptive')
    nlp.add_option('tol', 1e-7)

    #
    # Scale the problem (Just for demonstration purposes)
    #
    nlp.set_problem_scaling(
        obj_scaling=2,
        x_scaling=[1, 1, 1, 1]
        )
    nlp.add_option('nlp_scaling_method', 'user-scaling')

    #
    # Solve the problem
    #
    x, info = nlp.solve(x0)

    print("Solution of the primal variables: x=%s\n" % repr(x))

    print("Solution of the dual variables: lambda=%s\n" % repr(info['mult_g']))

    print("Objective=%s\n" % repr(info['obj_val']))


if __name__ == '__main__':
    try:
        main()
    except ValueError as e:
        assert str(e) == 'bum'
        print('Caught ValueError: {}'.format(e))



# -*- coding: utf-8 -*-
"""
cyipopt: Python wrapper for the Ipopt optimization package, written in Cython.

Copyright (C) 2012-2015 Amit Aides
Copyright (C) 2015-2017 Matthias Kümmerer
Copyright (C) 2017-2024 cyipopt developers

License: EPL 2.0
"""

class CyIpoptEvaluationError(ArithmeticError):
    """An exception that should be raised in evaluation callbacks to signal
    to CyIpopt that a numerical error occured during function evaluation.

    Whereas most exceptions that occur in callbacks are re-raised, exceptions
    of this type are ignored other than to communicate to Ipopt that an error
    occurred.

    Ipopt handles evaluation errors differently depending on where they are
    raised (which evaluation callback returns ``false`` to Ipopt).
    When evaluation errors are raised in the following callbacks, Ipopt
    attempts to recover by cutting the step size. This is usually the desired
    behavior when an undefined value is encountered.

    - ``objective``
    - ``constraints``

    When raised in the following callbacks, Ipopt fails with an "Invalid number"
    return status.

    - ``gradient``
    - ``jacobian``
    - ``hessian``

    Raising an evaluation error in the following callbacks results is not
    supported.

    - ``jacobianstructure``
    - ``hessianstructure``
    - ``intermediate``

    """

    pass



import numpy as np

import cyipopt

import pytest


pre_3_14_13 = (
    cyipopt.IPOPT_VERSION < (3, 14, 13)
)


def full_indices(shape):
    def indices():
        r, c = np.indices(shape)
        return r.flatten(), c.flatten()

    return indices


def tril_indices(size):
    def indices():
        return np.tril_indices(size)

    return indices


def flatten(func):
    def _func(*args):
        return func(*args).flatten()

    return _func


@pytest.fixture
def hs071_sparse_definition_fixture(hs071_variable_lower_bounds_fixture,
                                    hs071_constraint_lower_bounds_fixture,
                                    hs071_definition_instance_fixture):
    problem = hs071_definition_instance_fixture
    n = len(hs071_variable_lower_bounds_fixture)
    m = len(hs071_constraint_lower_bounds_fixture)

    problem.jacobianstructure = full_indices((m, n))
    problem.hessianstructure = tril_indices(n)

    problem.jacobian = flatten(problem.jacobian)
    problem.hessian = flatten(problem.hessian)

    return problem


@pytest.fixture
def hs071_sparse_instance(hs071_initial_guess_fixture,
                          hs071_variable_lower_bounds_fixture,
                          hs071_variable_upper_bounds_fixture,
                          hs071_constraint_lower_bounds_fixture,
                          hs071_constraint_upper_bounds_fixture,
                          hs071_sparse_definition_fixture):

    class Instance:
        pass

    instance = Instance()
    instance.problem_definition = hs071_sparse_definition_fixture
    instance.x0 = hs071_initial_guess_fixture
    instance.lb = hs071_variable_lower_bounds_fixture
    instance.ub = hs071_variable_upper_bounds_fixture
    instance.cl = hs071_constraint_lower_bounds_fixture
    instance.cu = hs071_constraint_upper_bounds_fixture
    instance.n = len(instance.x0)
    instance.m = len(instance.cl)

    return instance


def problem_for_instance(instance):
    return cyipopt.Problem(n=instance.n,
                           m=instance.m,
                           problem_obj=instance.problem_definition,
                           lb=instance.lb,
                           ub=instance.ub,
                           cl=instance.cl,
                           cu=instance.cu)


def test_solve_sparse(hs071_sparse_instance):
    instance = hs071_sparse_instance
    problem = problem_for_instance(instance)

    x, info = problem.solve(instance.x0)

    assert info['status'] == 0


def ensure_solve_status(instance, status):
    problem = problem_for_instance(instance)

    problem.add_option('max_iter', 50)
    x, info = problem.solve(instance.x0)

    assert info['status'] == status


def ensure_invalid_option(instance):
    # -12: Invalid option
    # Thrown in invalid Hessian because "hessian_approximation"
    # is not chosen as "limited-memory"
    ensure_solve_status(instance, -12)


def ensure_invalid_number(instance):
    # -13: Invalid Number Detected
    ensure_solve_status(instance, -13)


def ensure_unrecoverable_exception(instance):
    # -100: Unrecoverable Exception
    # *Should* be returned from errors in initialization
    ensure_solve_status(instance, -100)


@pytest.mark.skipif(pre_3_14_13, reason="Not caught in Ipopt < (3,14,13)")
def test_solve_neg_jac(hs071_sparse_instance):
    n = hs071_sparse_instance.n
    m = hs071_sparse_instance.m
    problem_definition = hs071_sparse_instance.problem_definition

    def jacobianstructure():
        r = np.full((m*n,), fill_value=-1, dtype=int)
        c = np.full((m*n,), fill_value=-1, dtype=int)
        return r, c

    problem_definition.jacobianstructure = jacobianstructure

    ensure_unrecoverable_exception(hs071_sparse_instance)


@pytest.mark.skipif(pre_3_14_13, reason="Not caught in Ipopt < (3,14,13)")
def test_solve_large_jac(hs071_sparse_instance):
    n = hs071_sparse_instance.n
    m = hs071_sparse_instance.m
    problem_definition = hs071_sparse_instance.problem_definition

    import logging
    logging.basicConfig(level=logging.DEBUG)

    def jacobianstructure():
        r = np.full((m*n,), fill_value=(m + n + 100), dtype=int)
        c = np.full((m*n,), fill_value=(m + n + 100), dtype=int)
        return r, c

    problem_definition.jacobianstructure = jacobianstructure

    ensure_unrecoverable_exception(hs071_sparse_instance)


@pytest.mark.skipif(pre_3_14_13, reason="Not caught in Ipopt < (3,14,13)")
def test_solve_wrong_jac_structure_size(hs071_sparse_instance):
    n = hs071_sparse_instance.n
    m = hs071_sparse_instance.m

    problem_definition = hs071_sparse_instance.problem_definition

    problem_definition.jacobianstructure = full_indices((m+1, n+1))

    ensure_unrecoverable_exception(hs071_sparse_instance)


@pytest.mark.skipif(pre_3_14_13, reason="Not caught in Ipopt < (3,14,13)")
def test_solve_wrong_jac_value_size(hs071_sparse_instance):
    n = hs071_sparse_instance.n
    m = hs071_sparse_instance.m

    problem_definition = hs071_sparse_instance.problem_definition

    def jacobian(x):
        return np.zeros((m*n + 10,))

    problem_definition.jacobian = jacobian

    ensure_invalid_number(hs071_sparse_instance)


def test_solve_triu_hess(hs071_sparse_instance):
    n = hs071_sparse_instance.n
    problem_definition = hs071_sparse_instance.problem_definition
    problem_definition.hessianstructure = lambda: np.triu_indices(n)

    ensure_invalid_option(hs071_sparse_instance)


def test_solve_neg_hess_entries(hs071_sparse_instance):
    n = hs071_sparse_instance.n
    problem_definition = hs071_sparse_instance.problem_definition

    def hessianstructure():
        r, c = np.tril_indices(n)
        rneg = np.full_like(r, -1, dtype=int)
        cneg = np.full_like(c, -1, dtype=int)
        return rneg, cneg

    problem_definition.hessianstructure = hessianstructure

    ensure_invalid_option(hs071_sparse_instance)


def test_solve_large_hess_entries(hs071_sparse_instance):
    n = hs071_sparse_instance.n
    problem_definition = hs071_sparse_instance.problem_definition

    def hessianstructure():
        r, c = np.tril_indices(n)
        rlarge = np.full_like(r, n + 100, dtype=int)
        clarge = np.full_like(c, n + 100, dtype=int)
        return rlarge, clarge

    problem_definition.hessianstructure = hessianstructure

    ensure_invalid_option(hs071_sparse_instance)


def test_solve_wrong_hess_struct_size(hs071_sparse_instance):
    n = hs071_sparse_instance.n
    problem_definition = hs071_sparse_instance.problem_definition

    def hessianstructure():
        return np.tril_indices(n + 10)

    problem_definition.hessianstructure = hessianstructure

    ensure_invalid_option(hs071_sparse_instance)


def test_solve_wrong_hess_value_size(hs071_sparse_instance):
    n = hs071_sparse_instance.n
    problem_definition = hs071_sparse_instance.problem_definition

    def hessian(x, lag, obj_factor):
        return np.zeros((n*n + 10,))

    problem_definition.hessian = hessian

    ensure_invalid_number(hs071_sparse_instance)



# -*- coding: utf-8 -*-
"""
cyipopt: Python wrapper for the Ipopt optimization package, written in Cython.

Copyright (C) 2012-2015 Amit Aides
Copyright (C) 2015-2017 Matthias Kümmerer
Copyright (C) 2017-2024 cyipopt developers

License: EPL 2.0
"""

import sys
import os.path
from distutils.sysconfig import get_python_lib
import subprocess as sp

from setuptools import setup
from setuptools.extension import Extension

# install requirements before import
from setuptools import dist
SETUP_REQUIRES = [
    "cython>=0.29.28",
    "numpy>=1.21.5",
    "setuptools>=44.1.1",
]
dist.Distribution().fetch_build_eggs(SETUP_REQUIRES)

from Cython.Distutils import build_ext
import numpy as np


exec(open("cyipopt/version.py", encoding="utf-8").read())
PACKAGE_NAME = "cyipopt"
DEPRECATED_PACKAGE_NAME = "ipopt"
VERSION = __version__
DESCRIPTION = "A Cython wrapper to the IPOPT optimization package"
with open("README.rst", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()
KEYWORDS = [
    "coin-or",
    "interior-point",
    "ipopt",
    "nlp",
    "nonlinear programming",
    "optimization",
]
AUTHOR = "Jason K. Moore"
EMAIL = "moorepants@gmail.com"
URL = "https://github.com/mechmotum/cyipopt"
INSTALL_REQUIRES = [
    "numpy>=1.21.5",
]
LICENSE = "EPL-2.0"
CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)",
    "Intended Audience :: Science/Research",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
]


def pkgconfig(*packages, **kw):
    """Returns a dictionary containing the include and library flags to pass to
    a Python extension that depends on the provided packages.

    Parameters
    ----------
    *packages : one or more strings
        These are the names of ``.pc`` files that pkg-config can locate, for
        example ``ipopt`` for the ``ipopt.pc`` file.
    **kw : list of strings
        Any values that should be preset in the returned dictionary. These can
        include lists of items for: ``include_dirs``, ``library_dirs``,
        ``libraries``, ``extra_compile_args``.

    Returns
    -------
    kw : dictionary
        Dictionary containing the keys: ``include_dirs``, ``library_dirs``,
        ``libraries``, ``extra_compile_args`` that are mapped to lists of
        strings.

    Notes
    -----
    This function is based on:

    http://code.activestate.com/recipes/502261-python-distutils-pkg-config/#c2

    """
    flag_map = {"-I": "include_dirs", "-L": "library_dirs", "-l": "libraries"}
    output = sp.Popen(["pkg-config", "--libs", "--cflags"] + list(packages),
                      stdout=sp.PIPE).communicate()[0]

    if not output:  # output will be empty string if pkg-config finds nothing
        msg = ("pkg-config was not able to find any of the requested packages "
               "{} on your system. Make sure pkg-config can discover the .pc "
               "files associated with the installed packages.")
        raise OSError(msg.format(list(packages)))

    output = output.decode("utf8")

    for token in output.split():
        if token[:2] in flag_map:
            kw.setdefault(flag_map.get(token[:2]), []).append(token[2:])
        else:
            kw.setdefault("extra_compile_args", []).append(token)

    kw["include_dirs"] += [np.get_include()]

    return kw


def handle_ext_modules_win_32_conda_forge_ipopt():
    conda_prefix = os.path.split(sys.executable)[0]

    IPOPT_INCLUDE_DIRS = [os.path.join(conda_prefix, "Library", "include",
                                       "coin-or"), np.get_include()]
    IPOPT_LIBS = ["ipopt-3"]
    IPOPT_LIB_DIRS = [os.path.join(conda_prefix, "Library", "lib")]
    EXT_MODULES = [Extension("ipopt_wrapper",
                             ["cyipopt/cython/ipopt_wrapper.pyx"],
                             include_dirs=IPOPT_INCLUDE_DIRS,
                             libraries=IPOPT_LIBS,
                             library_dirs=IPOPT_LIB_DIRS)]
    DATA_FILES = None
    include_package_data = True
    return EXT_MODULES, DATA_FILES, include_package_data


def handle_ext_modules_win_32_other_ipopt():
    IPOPT_INCLUDE_DIRS = [os.path.join(ipoptdir, "include", "coin-or"),
                          np.get_include()]

    # These are the specific binaries in the IPOPT 3.13.2 binary download:
    # https://github.com/coin-or/Ipopt/releases/download/releases%2F3.13.2/Ipopt-3.13.2-win64-msvs2019-md.zip
    IPOPT_LIBS = ["ipopt.dll", "ipoptamplinterface.dll"]
    IPOPT_LIB_DIRS = [os.path.join(ipoptdir, "lib")]

    IPOPT_DLL = [
        "ipopt-3.dll",
        "ipoptamplinterface-3.dll",
        "libifcoremd.dll",
        "libmmd.dll",
        "msvcp140.dll",
        "svml_dispmd.dll",
        "vcruntime140.dll",
    ]
    IPOPT_DLL_DIRS = [os.path.join(ipoptdir, "bin")]
    EXT_MODULES = [Extension("ipopt_wrapper",
                             ["cyipopt/cython/ipopt_wrapper.pyx"],
                             include_dirs=IPOPT_INCLUDE_DIRS,
                             libraries=IPOPT_LIBS,
                             library_dirs=IPOPT_LIB_DIRS)]
    DATA_FILES = [(get_python_lib(),
                  [os.path.join(IPOPT_DLL_DIRS[0], dll)
                   for dll in IPOPT_DLL])] if IPOPT_DLL else None
    include_package_data = False
    return EXT_MODULES, DATA_FILES, include_package_data


def handle_ext_modules_general_os():
    ipopt_wrapper_ext = Extension("ipopt_wrapper",
                                  ["cyipopt/cython/ipopt_wrapper.pyx"],
                                  **pkgconfig("ipopt"))
    EXT_MODULES = [ipopt_wrapper_ext]
    DATA_FILES = None
    include_package_data = True
    return EXT_MODULES, DATA_FILES, include_package_data


if __name__ == "__main__":

    ipoptdir = os.environ.get("IPOPTWINDIR", "")

    # conda-forge hosts a windows version of ipopt for ipopt versions >= 3.13.
    # The location of the headers and binaries are in $CONDA_PREFIX/Library/
    # and the library binary is named "libipopt.lib". If the IPOPTWINDIR
    # environment variable is set to USECONDAFORGEIPOPT then this setup will be
    # run.
    if sys.platform == "win32" and ipoptdir == "USECONDAFORGEIPOPT":
        print('Using Conda Forge Ipopt on Windows.')
        ext_module_data = handle_ext_modules_win_32_conda_forge_ipopt()
    elif sys.platform == "win32" and ipoptdir:
        print('Using Ipopt in {} directory on Windows.'.format(ipoptdir))
        ext_module_data = handle_ext_modules_win_32_other_ipopt()
    elif sys.platform == "win32" and not ipoptdir:
        ipoptdir = os.path.abspath(os.path.dirname(__file__))
        msg = 'Using Ipopt adjacent to setup.py in {} on Windows.'
        print(msg.format(ipoptdir))
        ext_module_data = handle_ext_modules_win_32_other_ipopt()
    else:
        print('Using Ipopt found with pkg-config.')
        ext_module_data = handle_ext_modules_general_os()
    EXT_MODULES, DATA_FILES, include_package_data = ext_module_data
    # NOTE : The `name` kwarg here is the distribution name, i.e. the name that
    # PyPi uses for a collection of packages. Historically this has been
    # `ipopt`, but as of 1.1.0 is `cyipopt`. `pip install cyipopt` will install
    # the `cyipopt` and `ipopt` packages into the `site-packages` directory.
    # Both `import cyipopt` and `import ipopt` will work, with the later giving
    # a deprecation warning.
    setup(name=PACKAGE_NAME,
          version=VERSION,
          author=AUTHOR,
          author_email=EMAIL,
          url=URL,
          description=DESCRIPTION,
          long_description=LONG_DESCRIPTION,
          keywords=KEYWORDS,
          license=LICENSE,
          classifiers=CLASSIFIERS,
          packages=[PACKAGE_NAME, DEPRECATED_PACKAGE_NAME],
          setup_requires=SETUP_REQUIRES,
          install_requires=INSTALL_REQUIRES,
          include_package_data=include_package_data,
          data_files=DATA_FILES,
          zip_safe=False,  # required for Py27 on Windows to work
          cmdclass={"build_ext": build_ext},
          ext_modules=EXT_MODULES,
          )



# -*- coding: utf-8 -*-
"""
cyipopt: Python wrapper for the Ipopt optimization package, written in Cython.

Copyright (C) 2012-2015 Amit Aides
Copyright (C) 2015-2017 Matthias Kümmerer
Copyright (C) 2017-2024 cyipopt developers

License: EPL 2.0
"""

import sys

import numpy as np
try:
    import scipy
except ImportError:  # scipy is not installed
    SCIPY_INSTALLED = False
else:
    SCIPY_INSTALLED = True
    del scipy
    from scipy import optimize
    import scipy.sparse
    try:
        from scipy.optimize import OptimizeResult
    except ImportError:
        # in scipy 0.14 Result was renamed to OptimizeResult
        from scipy.optimize import Result
        OptimizeResult = Result
    try:
        # MemoizeJac has been made a private class, see
        # https://github.com/scipy/scipy/issues/17572
        from scipy.optimize._optimize import MemoizeJac
    except ImportError:
        from scipy.optimize.optimize import MemoizeJac
    try:
        from scipy.sparse import coo_array
    except ImportError:
        # coo_array was introduced with scipy 1.8
        from scipy.sparse import coo_matrix as coo_array

import cyipopt


class IpoptProblemWrapper(object):
    """Class used to map a scipy minimize definition to a cyipopt problem.

    Parameters
    ==========
    fun : callable
        The objective function to be minimized: ``fun(x, *args, **kwargs) ->
        float``.
    args : tuple, optional
        Extra arguments passed to the objective function and its derivatives
        (``fun``, ``jac``, ``hess``).
    kwargs : dictionary, optional
        Extra keyword arguments passed to the objective function and its
        derivatives (``fun``, ``jac``, ``hess``).
    jac : callable, optional
        The Jacobian (gradient) of the objective function:
        ``jac(x, *args, **kwargs) -> ndarray, shape(n, )``.
        If ``None``, SciPy's ``approx_fprime`` is used.
    hess : callable, optional
        If ``None``, the Hessian is computed using IPOPT's numerical methods.
        Explicitly defined Hessians are not yet supported for this class.
    hessp : callable, optional
        If ``None``, the Hessian is computed using IPOPT's numerical methods.
        Explicitly defined Hessians are not yet supported for this class.
    constraints : {Constraint, dict} or List of {Constraint, dict}, optional
        See :py:func:`scipy.optimize.minimize` for more information. Note that
        the jacobian of each constraint corresponds to the `'jac'` key and must
        be a callable function with signature ``jac(x) -> {ndarray,
        coo_array}``. If the constraint's value of `'jac'` is a boolean and
        True, the constraint function `fun` is expected to return a tuple
        `(con_val, con_jac)` consisting of the evaluated constraint `con_val`
        and the evaluated jacobian `con_jac`.
    eps : float, optional
        Step size used in finite difference approximations of the objective
        function gradient and constraint Jacobian.
    con_dims : array_like, optional
        Dimensions p_1, ..., p_m of the m constraint functions
        g_1, ..., g_m : R^n -> R^(p_i).
    sparse_jacs: array_like, optional
        If sparse_jacs[i] = True, the i-th constraint's jacobian is sparse.
        Otherwise, the i-th constraint jacobian is assumed to be dense.
    jac_nnz_row: array_like, optional
        The row indices of the nonzero elements in the stacked
        constraint jacobian matrix
    jac_nnz_col: array_like, optional
        The column indices of the nonzero elements in the stacked
        constraint jacobian matrix
    """

    def __init__(self,
                 fun,
                 args=(),
                 kwargs=None,
                 jac=None,
                 hess=None,
                 hessp=None,
                 constraints=(),
                 eps=1e-8,
                 con_dims=(),
                 sparse_jacs=(),
                 jac_nnz_row=(),
                 jac_nnz_col=()):
        if not SCIPY_INSTALLED:
            msg = 'Install SciPy to use the `IpoptProblemWrapper` class.'
            raise ImportError()
        self.obj_hess = None
        self.last_x = None

        # Input validation of user-provided arguments
        if fun is not None and not callable(fun):
            raise ValueError('`fun` must be callable.')
        if not isinstance(args, tuple):
            args = (args,)
        kwargs = dict() if kwargs is None else kwargs
        if not isinstance(kwargs, dict):
            raise ValueError('`kwargs` must be a dictionary.')
        if jac is not None and jac not in {True, False} and not callable(jac):
            raise ValueError('`jac` must be callable or boolean.')
        if hess is not None and not callable(hess):
            raise ValueError('`hess` must be callable.')
        if hessp is not None:
            raise NotImplementedError(
                '`hessp` is not yet supported by Ipopt.`')
        # TODO: add input validation for `constraints` when adding
        #  support for instances of new-style constraints (e.g.
        #  `NonlinearConstraint`) and sequences of constraints.

        if hess is not None:
            self.obj_hess = hess
        if not jac:
            def jac(x, *args, **kwargs):
                def wrapped_fun(x):
                    return fun(x, *args, **kwargs)
                return optimize.approx_fprime(x, wrapped_fun, eps)
        elif jac is True:
            fun = MemoizeJac(fun)
            jac = fun.derivative

        self.fun = fun
        self.jac = jac
        self.args = args
        self.kwargs = kwargs or {}
        self._constraint_funs = []
        self._constraint_jacs = []
        self._constraint_hessians = []
        self._constraint_dims = np.asarray(con_dims)
        self._constraint_args = []
        self._constraint_kwargs = []
        self._constraint_jac_is_sparse = sparse_jacs
        self._constraint_jacobian_structure = (jac_nnz_row, jac_nnz_col)
        if isinstance(constraints, dict):
            constraints = (constraints, )
        for con in constraints:
            con_fun = con['fun']
            con_jac = con.get('jac', None)
            con_args = con.get('args', [])
            con_hessian = con.get('hess', None)
            con_kwargs = con.get('kwargs', {})
            if con_jac is None:
                # beware of late binding!
                def con_jac(x, *args, con_fun=con_fun, **kwargs):
                    def wrapped(x):
                        return con_fun(x, *args, **kwargs)
                    return optimize.approx_fprime(x, wrapped, eps)
            elif con_jac is True:
                con_fun = MemoizeJac(con_fun)
                con_jac = con_fun.derivative
            elif not callable(con_jac):
                raise NotImplementedError('jac has to be bool or a function')
            if (self.obj_hess is not None
                    and con_hessian is None) or (self.obj_hess is None
                                                 and con_hessian is not None):
                msg = "hessian has to be provided for the objective and all constraints"
                raise NotImplementedError(msg)
            self._constraint_funs.append(con_fun)
            self._constraint_jacs.append(con_jac)
            self._constraint_hessians.append(con_hessian)
            self._constraint_args.append(con_args)
            self._constraint_kwargs.append(con_kwargs)
        # Set up evaluation counts
        self.nfev = 0
        self.njev = 0
        self.nit = 0

    def evaluate_fun_with_grad(self, x):
        """ For backwards compatibility. """
        return (self.objective(x), self.gradient(x, **self.kwargs))

    def objective(self, x):
        self.nfev += 1
        return self.fun(x, *self.args, **self.kwargs)

    # TODO : **kwargs is ignored, not sure why it is here.
    def gradient(self, x, **kwargs):
        self.njev += 1
        return self.jac(x, *self.args, **self.kwargs)  # .T

    def constraints(self, x):
        con_values = []
        for fun, args, kwargs in zip(self._constraint_funs,
                                     self._constraint_args,
                                     self._constraint_kwargs):
            con_values.append(fun(x, *args, **kwargs))
        return np.hstack(con_values)

    def jacobianstructure(self):
        return self._constraint_jacobian_structure

    def jacobian(self, x):
        # Convert all dense constraint jacobians to sparse ones.
        # The structure ( = row and column indices) is already known at this point,
        # so we only need to stack the evaluated jacobians
        jac_values = []
        for i, (jac, args, kwargs) in enumerate(zip(self._constraint_jacs,
                                                    self._constraint_args,
                                                    self._constraint_kwargs)):
            if self._constraint_jac_is_sparse[i]:
                jac_val = jac(x, *args, **kwargs)
                jac_values.append(jac_val.data)
            else:
                dense_jac_val = np.atleast_2d(jac(x, *args, **kwargs))
                jac_values.append(dense_jac_val.ravel())
        return np.hstack(jac_values)

    def hessian(self, x, lagrange, obj_factor):
        H = obj_factor * self.obj_hess(x, *self.args, **self.kwargs)  # type: ignore
        # split the lagrangian multipliers for each constraint hessian
        lagrs = np.split(lagrange, np.cumsum(self._constraint_dims[:-1]))
        for hessian, args, kwargs, lagr in zip(self._constraint_hessians,
                                               self._constraint_args,
                                               self._constraint_kwargs, lagrs):
            H += hessian(x, lagr, *args, **kwargs)
        return H[np.tril_indices(x.size)]

    def intermediate(self, alg_mod, iter_count, obj_value, inf_pr, inf_du, mu,
                     d_norm, regularization_size, alpha_du, alpha_pr,
                     ls_trials):

        self.nit = iter_count


def get_bounds(bounds):
    if bounds is None:
        return None, None
    else:
        lb = [b[0] for b in bounds]
        ub = [b[1] for b in bounds]
        return lb, ub


def _get_sparse_jacobian_structure(constraints, x0):
    con_jac_is_sparse = []
    jacobians = []
    x0 = np.asarray(x0)
    if isinstance(constraints, dict):
        constraints = (constraints, )
    if len(constraints) == 0:
        return [], [], []
    for con in constraints:
        con_jac = con.get('jac', False)
        if con_jac:
            if isinstance(con_jac, bool):
                _, jac_val = con['fun'](x0, *con.get('args', []),
                                        **con.get('kwargs', {}))
            else:
                jac_val = con_jac(x0, *con.get('args', []),
                                  **con.get('kwargs', {}))
            # check if dense or sparse
            if isinstance(jac_val, coo_array):
                jacobians.append(jac_val)
                con_jac_is_sparse.append(True)
            else:
                # Creating the coo_array from jac_val would yield to
                # wrong dimensions if some values in jac_val are zero,
                # so we assume all values in jac_val are nonzero
                jacobians.append(coo_array(np.ones_like(np.atleast_2d(jac_val))))
                con_jac_is_sparse.append(False)
        else:
            # we approximate this jacobian later (=dense)
            con_val = np.atleast_1d(con['fun'](x0, *con.get('args', []),
                                               **con.get('kwargs', {})))
            jacobians.append(coo_array(np.ones((con_val.size, x0.size))))
            con_jac_is_sparse.append(False)
    J = scipy.sparse.vstack(jacobians)
    return con_jac_is_sparse, J.row, J.col


def get_constraint_dimensions(constraints, x0):
    con_dims = []
    if isinstance(constraints, dict):
        constraints = (constraints, )
    for con in constraints:
        if con.get('jac', False) is True:
            m = len(np.atleast_1d(con['fun'](x0, *con.get('args', []),
                                             **con.get('kwargs', {}))[0]))
        else:
            m = len(np.atleast_1d(con['fun'](x0, *con.get('args', []),
                                             **con.get('kwargs', {}))))
        con_dims.append(m)
    return np.array(con_dims)


def get_constraint_bounds(constraints, x0, INF=1e19):
    cl = []
    cu = []
    if isinstance(constraints, dict):
        constraints = (constraints, )
    for con in constraints:
        if con.get('jac', False) is True:
            m = len(np.atleast_1d(con['fun'](x0, *con.get('args', []),
                                             **con.get('kwargs', {}))[0]))
        else:
            m = len(np.atleast_1d(con['fun'](x0, *con.get('args', []),
                                             **con.get('kwargs', {}))))
        cl.extend(np.zeros(m))
        if con['type'] == 'eq':
            cu.extend(np.zeros(m))
        elif con['type'] == 'ineq':
            cu.extend(INF * np.ones(m))
        else:
            raise ValueError(con['type'])
    cl = np.array(cl)
    cu = np.array(cu)

    return cl, cu


def replace_option(options, oldname, newname):
    if oldname in options:
        if newname not in options:
            options[newname] = options.pop(oldname)


def convert_to_bytes(options):
    if sys.version_info >= (3, 0):
        for key in list(options.keys()):
            try:
                if bytes(key, 'utf-8') != key:
                    options[bytes(key, 'utf-8')] = options[key]
                    options.pop(key)
            except TypeError:
                pass


def _wrap_fun(fun, kwargs):
    if callable(fun) and kwargs:
        def new_fun(x, *args):
            return fun(x, *args, **kwargs)
    else:
        new_fun = fun
    return new_fun

def _wrap_funs(fun, jac, hess, hessp, constraints, kwargs):
    wrapped_fun = _wrap_fun(fun, kwargs)
    wrapped_jac = _wrap_fun(jac, kwargs)
    wrapped_hess = _wrap_fun(hess, kwargs)
    wrapped_hessp = _wrap_fun(hessp, kwargs)
    if isinstance(constraints, dict):
        constraints = (constraints,)
    wrapped_constraints = []
    for constraint in constraints:
        constraint = constraint.copy()
        ckwargs = constraint.pop('kwargs', {})
        constraint['fun'] = _wrap_fun(constraint.get('fun', None), ckwargs)
        constraint['jac'] = _wrap_fun(constraint.get('jac', None), ckwargs)
        wrapped_constraints.append(constraint)
    return (wrapped_fun, wrapped_jac, wrapped_hess, wrapped_hessp,
            wrapped_constraints)



def minimize_ipopt(fun,
                   x0,
                   args=(),
                   kwargs=None,
                   method=None,
                   jac=None,
                   hess=None,
                   hessp=None,
                   bounds=None,
                   constraints=(),
                   tol=None,
                   callback=None,
                   options=None):
    """
    Minimization using Ipopt with an interface like
    :py:func:`scipy.optimize.minimize`.

    Differences compared to :py:func:`scipy.optimize.minimize` include:

    - A different default `method`: when `method` is not provided, Ipopt is
      used to solve the problem.
    - Support for parameter `kwargs`: additional keyword arguments to be
      passed to the objective function, constraints, and their derivatives.
    - Lack of support for `callback` and `hessp` with the default `method`.

    This function can be used to solve general nonlinear programming problems
    of the form:

    .. math::

       \min_ {x \in R^n} f(x)

    subject to

    .. math::

       g_L \leq g(x) \leq g_U

       x_L \leq  x  \leq x_U

    where :math:`x` are the optimization variables, :math:`f(x)` is the
    objective function, :math:`g(x)` are the general nonlinear constraints,
    and :math:`x_L` and :math:`x_U` are the upper and lower bounds
    (respectively) on the decision variables. The constraints, :math:`g(x)`,
    have lower and upper bounds :math:`g_L` and :math:`g_U`. Note that equality
    constraints can be specified by setting :math:`g^i_L = g^i_U`.

    Parameters
    ----------
    fun : callable
        The objective function to be minimized: ``fun(x, *args, **kwargs) ->
        float``.
    x0 : array-like, shape(n, )
        Initial guess. Array of real elements of shape (n,),
        where ``n`` is the number of independent variables.
    args : tuple, optional
        Extra arguments passed to the objective function and its
        derivatives (``fun``, ``jac``, and ``hess``).
    kwargs : dictionary, optional
        Extra keyword arguments passed to the objective function and its
        derivatives (``fun``, ``jac``, ``hess``).
    method : str, optional
        If unspecified (default), Ipopt is used.
        :py:func:`scipy.optimize.minimize` methods can also be used.
    jac : callable, optional
        The Jacobian (gradient) of the objective function:
        ``jac(x, *args, **kwargs) -> ndarray, shape(n, )``.
        If ``None``, SciPy's ``approx_fprime`` is used.
    hess : callable, optional
        The Hessian of the objective function:
        ``hess(x) -> ndarray, shape(n, )``.
        If ``None``, the Hessian is computed using IPOPT's numerical methods.
    hessp : callable, optional
        If `method` is one of the SciPy methods, this is a callable that
        produces the inner product of the Hessian and a vector. Otherwise, an
        error will be raised if a value other than ``None`` is provided.
    bounds :  sequence of shape(n, ) or :py:class:`scipy.optimize.Bounds`, optional
        Simple bounds on decision variables. There are two ways to specify the
        bounds:

            1. Instance of :py:class:`scipy.optimize.Bounds` class.
            2. Sequence of ``(min, max)`` pairs for each element in `x`. Use
               ``None`` to specify an infinite bound (i.e., no bound).

    constraints : {Constraint, dict}, optional
        Linear or nonlinear constraint specified by a dictionary,
        :py:class:`scipy.optimize.LinearConstraint`, or
        :py:class:`scipy.optimize.NonlinearConstraint`.
        See :py:func:`scipy.optimize.minimize` for more information. Note that
        the Jacobian of each constraint corresponds to the ``'jac'`` key and
        must be a callable function with signature ``jac(x) -> {ndarray,
        coo_array}``. If the constraint's value of ``'jac'`` is ``True``, the
        constraint function ``fun`` must return a tuple ``(con_val, con_jac)``
        consisting of the evaluated constraint ``con_val`` and the evaluated
        Jacobian ``con_jac``.
    tol : float, optional (default=1e-8)
        The desired relative convergence tolerance, passed as an option to
        Ipopt. See [1]_ for details.
    options : dict, optional
        A dictionary of solver options.

        When `method` is unspecified (default: Ipopt), the options
        ``disp`` and ``maxiter`` are automatically mapped to their Ipopt
        equivalents ``print_level`` and ``max_iter``, and ``eps`` is used to
        control the step size of finite difference gradient and constraint
        Jacobian approximations. All other options are passed directly
        to Ipopt. See [1]_ for details.

        For other values of `method`, `options` is passed to the SciPy solver.
        See [2]_ for details.
    callback : callable, optional
        This argument is ignored by the default `method` (Ipopt).
        If `method` is one of the SciPy methods, this is a callable that is
        called once per iteration. See [2]_ for details.

    References
    ----------
    .. [1] COIN-OR Project. "Ipopt: Ipopt Options".
           https://coin-or.github.io/Ipopt/OPTIONS.html
    .. [2] The SciPy Developers. "scipy.optimize.minimize".
           https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html

    Examples
    --------
    Consider the problem of minimizing the Rosenbrock function. The Rosenbrock
    function and its derivatives are implemented in
    :py:func:`scipy.optimize.rosen`, :py:func:`scipy.optimize.rosen_der`, and
    :py:func:`scipy.optimize.rosen_hess`.

    >>> from cyipopt import minimize_ipopt
    >>> from scipy.optimize import rosen, rosen_der
    >>> x0 = [1.3, 0.7, 0.8, 1.9, 1.2]  # initial guess

    If we provide the objective function but no derivatives, Ipopt finds the
    correct minimizer (``[1, 1, 1, 1, 1]``) with a minimum objective value of
    0. However, it does not report success, and it requires many iterations
    and function evaluations before termination. This is because SciPy's
    ``approx_fprime`` requires many objective function evaluations to
    approximate the gradient, and still the approximation is not very accurate,
    delaying convergence.

    >>> res = minimize_ipopt(rosen, x0, jac=rosen_der)
    >>> res.success
    False
    >>> res.x
    array([1., 1., 1., 1., 1.])
    >>> res.nit, res.nfev, res.njev
    (46, 528, 48)

    To improve performance, provide the gradient using the `jac` keyword.
    In this case, Ipopt recognizes its own success, and requires fewer function
    evaluations to do so.

    >>> res = minimize_ipopt(rosen, x0, jac=rosen_der)
    >>> res.success
    True
    >>> res.nit, res.nfev, res.njev
    (37, 200, 39)

    For best results, provide the Hessian, too.

    >>> res = minimize_ipopt(rosen, x0, jac=rosen_der, hess=rosen_hess)
    >>> res.success
    True
    >>> res.nit, res.nfev, res.njev
    (17, 29, 19)
    """
    if not SCIPY_INSTALLED:
        msg = 'Install SciPy to use the `minimize_ipopt` function.'
        raise ImportError(msg)

    res = _minimize_ipopt_iv(fun, x0, args, kwargs, method, jac, hess, hessp,
                             bounds, constraints, tol, callback, options)
    (fun, x0, args, kwargs, method, jac, hess, hessp,
     bounds, constraints, tol, callback, options) = res

    if method is not None:
        funs = _wrap_funs(fun, jac, hess, hessp, constraints, kwargs)
        fun, jac, hess, hessp, constraints = funs
        bounds = optimize.Bounds(*bounds)
        res = optimize.minimize(fun, x0, args, method, jac, hess, hessp,
                                bounds, constraints, tol, callback, options)
        return res

    _x0 = np.atleast_1d(x0)

    lb, ub = bounds
    cl, cu = get_constraint_bounds(constraints, _x0)
    con_dims = get_constraint_dimensions(constraints, _x0)
    sparse_jacs, jac_nnz_row, jac_nnz_col = _get_sparse_jacobian_structure(
        constraints, x0)

    if options is None:
        options = {}
    eps = options.pop('eps', 1e-8)

    problem = IpoptProblemWrapper(fun,
                                  args=args,
                                  kwargs=kwargs,
                                  jac=jac,
                                  hess=hess,
                                  hessp=hessp,
                                  constraints=constraints,
                                  eps=eps,
                                  con_dims=con_dims,
                                  sparse_jacs=sparse_jacs,
                                  jac_nnz_row=jac_nnz_row,
                                  jac_nnz_col=jac_nnz_col)

    nlp = cyipopt.Problem(n=len(x0),
                          m=len(cl),
                          problem_obj=problem,
                          lb=lb,
                          ub=ub,
                          cl=cl,
                          cu=cu)

    # python3 compatibility
    convert_to_bytes(options)

    # Rename some default scipy options
    replace_option(options, b'disp', b'print_level')
    replace_option(options, b'maxiter', b'max_iter')
    options[b'print_level'] = int(options.get(b'print_level', 0))

    if b'tol' not in options:
        options[b'tol'] = tol or 1e-8
    if b'mu_strategy' not in options:
        options[b'mu_strategy'] = b'adaptive'
    if b'hessian_approximation' not in options:
        if hess is None and hessp is None:
            options[b'hessian_approximation'] = b'limited-memory'
    for option, value in options.items():
        try:
            nlp.add_option(option, value)
        except TypeError as e:
            msg = 'Invalid option for IPOPT: {0}: {1} (Original message: "{2}")'
            raise TypeError(msg.format(option, value, e))

    x, info = nlp.solve(x0)

    return OptimizeResult(x=x,
                          success=info['status'] == 0,
                          status=info['status'],
                          message=info['status_msg'],
                          fun=info['obj_val'],
                          info=info,
                          nfev=problem.nfev,
                          njev=problem.njev,
                          nit=problem.nit)


def _minimize_ipopt_iv(fun, x0, args, kwargs, method, jac, hess, hessp,
                       bounds, constraints, tol, callback, options):
    # basic input validation for minimize_ipopt that is not included in
    # IpoptProblemWrapper
    x0 = np.atleast_1d(x0)
    if not np.issubdtype(x0.dtype, np.number):
        raise ValueError('`x0` must be a numeric array.')

    # `method` does not need input validation. If `method is not None`, it is
    # passed to `scipy.optimize.minimize`, which raises a readable error if
    # the value isn't recognized.

    # Handle bounds that are either sequences (of sequences) or instances of
    # `optimize.Bounds`.
    if bounds is None:
        bounds = [-np.inf, np.inf]

    if isinstance(bounds, optimize.Bounds):
        lb, ub = bounds.lb, bounds.ub
    else:
        bounds = np.atleast_2d(bounds)
        if bounds.shape[1] != 2:
            raise ValueError("`bounds` must specify both lower and upper "
                             "limits for each decision variable.")
        lb, ub = bounds.T

    try:
        lb, ub, x0 = np.broadcast_arrays(lb, ub, x0)
    except ValueError:
        raise ValueError("The number of lower bounds, upper bounds, and "
                         "decision variables must be equal or broadcastable.")

    try:
        lb = lb.astype(np.float64)
        ub = ub.astype(np.float64)
    except ValueError:
        raise ValueError("The bounds must be numeric.")

    # Nones turn into NaNs above. Previously, NaNs caused Ipopt to hang, so
    # I'm not concerned about turning them into infs.
    lb[np.isnan(lb)] = -np.inf
    ub[np.isnan(ub)] = np.inf

    bounds = lb, ub

    constraints = optimize._minimize.standardize_constraints(constraints, x0,
                                                             'old')

    if method is None and callback is not None:
        raise NotImplementedError('`callback` is not yet supported by Ipopt.`')

    if tol is not None:
        tol = np.asarray(tol)[()]
        if tol.ndim != 0 or not np.issubdtype(tol.dtype, np.number) or tol <= 0:
            raise ValueError('`tol` must be a positive scalar.')
        else:  # tol should be a float, not an array
            tol = float(tol)

    options = dict() if options is None else options
    if not isinstance(options, dict):
        raise ValueError('`options` must be a dictionary.')

    return (fun, x0, args, kwargs, method, jac, hess, hessp,
            bounds, constraints, tol, callback, options)
